/**
 * Architektur & Implementierung von DBMS, Aufgabenblatt 7
 *
 * Bitmap Indexing, Query Processing and WAH Compression
 *
 * File        : WAHBitVector.h
 */

#ifndef WAHBITVECTOR_H
#define WAHBITVECTOR_H

#include <stdint.h>
#include <vector>
#include <bitset>
#include "PlainBitVector.h"
#include <iostream>

class WAHBitVector;

/***** Global bit operations on WAHBitVectors *****/
WAHBitVector operator& (const WAHBitVector& lhs, const WAHBitVector& rhs);
WAHBitVector operator| (const WAHBitVector& lhs, const WAHBitVector& rhs);


/***** WAHBitVector class represents compressed bit vectors with WAH *****/
class WAHBitVector {
  private:
    uint32_t mSize;
    std::vector < std::bitset <32> > data;
    PlainBitVector decompressed;
    
    uint32_t mNumWords;
    uint32_t writeSetIdx;
    uint32_t last_idx;
  public:
    WAHBitVector(uint32_t size) 
        : mSize(size), decompressed(0)
      {
        mNumWords=2;
        data.resize(mNumWords);
        last_idx = 0;
        writeSetIdx = 0;
      }
   
    PlainBitVector getDecompressed() const {
        if(decompressed.size() > 0) return decompressed;
        PlainBitVector decoded(mSize);
        uint32_t idx = 0;
        for(std::bitset<32> set : data) {
            if(idx >= mSize) break;
            uint32_t coded = set.to_ulong();
            if(coded > 0x80000000) {
                idx += (coded - 0x80000000) * 31;
            } else {
                for(int i=0; i<31; i++, idx++) {
                    if(set[i]) { 
                        decoded.nextSetBitIndex(idx);
                     }
                }
            }
        }
        return decoded;
    }

 
    void decompress() {
        if(decompressed.size() == 0) {
          this->decompressed = getDecompressed();
        }
    }

    /** number of bits */
    uint32_t size() const {
        return mSize;
    }

    /** number of 32-bit words required to store the bit vector */
    uint32_t getNumWords() const {
        return mNumWords;
    }

    friend WAHBitVector operator& (const WAHBitVector& lhs, const WAHBitVector& rhs);
    friend WAHBitVector operator| (const WAHBitVector& lhs, const WAHBitVector& rhs);

    /**
     * \defgroup BitOperations Bit operation methods over the compressed bit vector.
     * @{ 
     */
    /**
     * In-place AND operator. Assumption vector lengths will be same. In-place
     * in the sense that the result will still be stored in this object although
     * internal representation is highly-likely to change, may even store the
     * result uncompressed.
     */
    WAHBitVector & operator&= (const WAHBitVector& rhs) {
        decompress();
        decompressed &= rhs.getDecompressed();
        return *this;
    }

    /* In-place OR operator. Assumption vector lengths will be same. */
    WAHBitVector & operator|= (const WAHBitVector& rhs) {
        decompress();
        decompressed |= rhs.getDecompressed();
        return *this;
    }

    /** Count set bits in a bit vector */
    uint32_t count() {
        decompress();
        return decompressed.count();
    }
    /**@}*/



    std::bitset<32> genEmpty(uint32_t n) {
        uint32_t code = n;
        code |= 0x80000000;
        return std::bitset<32>(code);
    }
    
    /** 
     * When building the bitvector, this method is called to inform about the
     * next set-bit's (bit with value 1) index.
     *
     * Subsequent calls of this method must have a greater idx than the previous
     * call. 
     * 
     * @param idx Index of the next set-bit in the bitvector.
     */
    void nextSetBitIndex(uint32_t idx) {

        // expand allocation
        if(writeSetIdx + 2 >= mNumWords) {
            mNumWords += 8;
            data.resize(mNumWords);
        }

        // compute gap
        uint32_t lastSetIdx = last_idx / 31; 
        uint32_t setIdx = idx / 31;
        int gap = setIdx - lastSetIdx - 1;
        
        // last_idx and idx in different bitsets
        if(lastSetIdx < setIdx) {
            writeSetIdx++;
        }     
       
        // fill gap with encoded zeros
        if(gap > 0) {
            data[writeSetIdx] = genEmpty(gap);
            writeSetIdx++;
        }

        // write bit
        data[writeSetIdx][idx%31] = 1;
        // remember writepos
        last_idx = idx;
    }

    /**
     * \defgroup BitIteration Iteration methods over the bits of the bit vector.
     * @{ 
     */
    void begin() {
        decompress();
        decompressed.begin();
    }

    /** has more bits? */
    bool hasNext() {
        return decompressed.hasNext();
    }

    /** return next bit, 0:false, 1:true */
    bool getNext() {
        return decompressed.getNext();
    }
    /**@}*/
};

#endif /* BITVECTOR_H */
