/**
 * Architektur & Implementierung von DBMS, Aufgabenblatt 7
 *
 * Bitmap Indexing, Query Processing and WAH Compression
 */
#if defined(__i386__)
#define _FILE_OFFSET_BITS 64
#endif

#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "schema.h"
#include "load.h"

/* max. number of characters in a .tbl input file */
#define MAX_LINE_LENGTH 256     

/* size of a memory allocation unit in bytes */
#define MEMORY_BLOCK    (100*1024*1024)

/*
 * helper macros for input file parsing 
 */

#define PARSE_IDENTIFIER(string, field) \
   field = atoi(string); \
   string = strtok(NULL, "|") 
#define PARSE_INTEGER(string, field) \
   field = atoi(string); \
   string = strtok(NULL, "|") 
#define PARSE_DECIMAL(string, field) \
   {  \
      char *s = strstr(string, "."); \
      if (s == NULL) {  \
         field = atoll(string)*100LL; \
      } else { \
         field = atoll(string)*100LL + atoi(++s); \
      } \
      string = strtok(NULL, "|"); \
   }
#define PARSE_CHAR(string, field) \
   field = string[0]; \
   string = strtok(NULL, "|") 
#define PARSE_TEXT_ROW(string, row, length, field) \
   strncpy(row.field, string, length); \
   row.field[length-1] = '\0'; \
   string = strtok(NULL, "|")
#define PARSE_TEXT_COLUMN(string, row, length, field) \
   strncpy(&(field[row*length]), string, length); \
   field[row*length+length-1] = '\0'; \
   string = strtok(NULL, "|")
#define PARSE_DATE(string, field) \
   { \
      uint32_t year, month, day; \
      char *s = strstr(string, "-"); \
      if (s == NULL) {  \
         fprintf(stderr, "invalid date string \"%s\"\n", string); \
         break; \
      } \
      year = atoi(string); \
      month = atoi(++s); \
      s = strstr(s, "-"); \
      if (s == NULL) {  \
         fprintf(stderr, "invalid date string \"%s\"\n", str);   \
         break;   \
      }  \
      day = atoi(s+1); \
      field = 10000*year + 100 * month + day;  \
      string = strtok(NULL, "|"); \
   }
#define ALLOCATE_COLUMN(column, type) \
   *column = (type*)malloc(MEMORY_BLOCK); \
   if (*column == NULL) { \
      fprintf(stderr, "unable to allocate memory for "#type); \
      return -1; \
   } 
#define EXPAND_BUFFER(column, name, type, elementsize) \
   if (currsize[name] < (count+1)*elementsize) {   \
      currsize[name] += MEMORY_BLOCK; \
      *column = (type*)realloc(*column, currsize[name]); \
      if (*column == NULL) { \
         fprintf(stderr, "unable to extend memory for column "#column"\n"); \
      } \
   }


/* load data file into row structure 
 *
 * \param filename  name .tbl file to read (input)
 * \param firstrow  pointer to first row in row array (output)
 *                  memory will be allocated in this function
 * \param numrows   number of rows loaded (output)
 * \return zero upon success, error code otherwise
 */
int 
load_lineitems_rows(char *filename, lineitem_t **firstrow, uint32_t *numrows)
{
   lineitem_t *rows;
   uint32_t    count = 0, currsize = MEMORY_BLOCK;
   char        buffer[MAX_LINE_LENGTH+1];
   char        *str, *tok;
   FILE *f;

   if ((f=fopen(filename, "r"))==NULL) { /* open data file */
      perror("cannot open file"); 
      return -1;
   }
   rows = (lineitem_t*)malloc(MEMORY_BLOCK); /* allocate first memory block */
   if (rows == NULL) {
      perror("unable to allocate memory"); 
      fclose(f);
      return -1;
   }

   str = fgets(buffer, sizeof(buffer), f); 
   while (str != NULL) {
      tok = strtok(buffer, "|"); 
      PARSE_IDENTIFIER(tok, rows[count].orderkey); 
      PARSE_IDENTIFIER(tok, rows[count].partkey); 
      PARSE_IDENTIFIER(tok, rows[count].suppkey); 
      PARSE_INTEGER(tok, rows[count].linenumber); 
      PARSE_INTEGER(tok, rows[count].quantity); 
      PARSE_DECIMAL(tok, rows[count].extendedprice); 
      PARSE_DECIMAL(tok, rows[count].discount); 
      PARSE_DECIMAL(tok, rows[count].tax); 
      PARSE_CHAR(tok, rows[count].returnflag); 
      PARSE_CHAR(tok, rows[count].linestatus); 
      PARSE_DATE(tok, rows[count].shipdate); 
      PARSE_DATE(tok, rows[count].commitdate); 
      PARSE_DATE(tok, rows[count].receiptdate); 
      PARSE_TEXT_ROW(tok, rows[count], 26, shipinstruct); 
      PARSE_TEXT_ROW(tok, rows[count], 11, shipmode); 
      PARSE_TEXT_ROW(tok, rows[count], 45, comment); 
      count++;
      if ((count+1)*sizeof(lineitem_t) > currsize) {
         /* next row would not fit in memory block */
         rows = (lineitem_t*) realloc(rows, currsize+MEMORY_BLOCK); 
         if (rows == NULL) {
            perror("unable to reallocate (increase) memory"); 
            free(rows); 
            fclose(f); 
            return -1;
         }
         currsize += MEMORY_BLOCK; 
      }

      str = fgets(buffer, sizeof(buffer), f); 
   }

   fclose(f);
   *firstrow = rows; 
   *numrows = count;
   return 0;
}


/* load data file into column arrays
 *
 * \param filename  name .tbl file to read (input)
 * \param numrows   number of rows loaded (output)
 * \param columsn   pointer to first row in each column array (output)
 *                  memory will be allocated in this function
 * \return zero upon success, error code otherwise
 */
int 
load_lineitems_columns(char *filename, uint32_t *numrows,
         uint32_t **orderkey, uint32_t **partkey, uint32_t **suppkey, 
         int32_t **linenumber, int64_t **quantity, int64_t **extendedprice,
         int64_t **discount, int64_t **tax, char **returnflag, 
         char **linestatus, uint32_t **shipdate, uint32_t **commitdate, 
         uint32_t **receiptdate, char **shipinstruct, 
         char **shipmode, char **comment)
{
   uint32_t count = 0, currsize[NUM_ATTRIBUTES], i; 
   char     buffer[MAX_LINE_LENGTH+1];
   char     *str, *tok;
   FILE     *f;

   if ((f=fopen(filename, "r"))==NULL) { /* open data file */
      perror("cannot open file"); 
      return -1;
   }
   for (i=0; i<NUM_ATTRIBUTES; i++) {
      currsize[i] = MEMORY_BLOCK;
   }
   ALLOCATE_COLUMN(orderkey, uint32_t);
   ALLOCATE_COLUMN(partkey, uint32_t);
   ALLOCATE_COLUMN(suppkey, uint32_t);
   ALLOCATE_COLUMN(linenumber, int32_t);
   ALLOCATE_COLUMN(quantity, int64_t);
   ALLOCATE_COLUMN(extendedprice, int64_t);
   ALLOCATE_COLUMN(discount, int64_t);
   ALLOCATE_COLUMN(tax, int64_t);
   ALLOCATE_COLUMN(returnflag, char);
   ALLOCATE_COLUMN(linestatus, char);
   ALLOCATE_COLUMN(shipdate, uint32_t);
   ALLOCATE_COLUMN(commitdate, uint32_t);
   ALLOCATE_COLUMN(receiptdate, uint32_t);
   ALLOCATE_COLUMN(shipinstruct, char); 
   ALLOCATE_COLUMN(shipmode, char);
   ALLOCATE_COLUMN(comment, char);

   str = fgets(buffer, sizeof(buffer), f); 
   while (str != NULL) {
      tok = strtok(buffer, "|"); 
      PARSE_IDENTIFIER(tok, (*orderkey)[count]); 
      PARSE_IDENTIFIER(tok, (*partkey)[count]); 
      PARSE_IDENTIFIER(tok, (*suppkey)[count]); 
      PARSE_INTEGER(tok, (*linenumber)[count]); 
      PARSE_INTEGER(tok, (*quantity)[count]);
      PARSE_DECIMAL(tok, (*extendedprice)[count]); 
      PARSE_DECIMAL(tok, (*discount)[count]); 
      PARSE_DECIMAL(tok, (*tax)[count]); 
      PARSE_CHAR(tok, (*returnflag)[count]);
      PARSE_CHAR(tok, (*linestatus)[count]);
      PARSE_DATE(tok, (*shipdate)[count]);
      PARSE_DATE(tok, (*commitdate)[count]);
      PARSE_DATE(tok, (*receiptdate)[count]);
      PARSE_TEXT_COLUMN(tok, count, 26, (*shipinstruct)); 
      PARSE_TEXT_COLUMN(tok, count, 11, (*shipmode)); 
      PARSE_TEXT_COLUMN(tok, count, 45, (*comment));
      count++;
   
      EXPAND_BUFFER(orderkey, L_ORDERKEY, uint32_t, sizeof(uint32_t));
      EXPAND_BUFFER(partkey, L_PARTKEY, uint32_t, sizeof(uint32_t));
      EXPAND_BUFFER(suppkey, L_SUPPKEY, uint32_t, sizeof(uint32_t));
      EXPAND_BUFFER(linenumber, L_LINENUMBER, int32_t, sizeof(int32_t));
      EXPAND_BUFFER(quantity, L_QUANTITY, int64_t, sizeof(int64_t));
      EXPAND_BUFFER(extendedprice, L_EXTENDEDPRICE, int64_t, sizeof(int64_t));
      EXPAND_BUFFER(discount, L_DISCOUNT, int64_t, sizeof(int64_t));
      EXPAND_BUFFER(tax, L_TAX, int64_t, sizeof(int64_t));
      EXPAND_BUFFER(returnflag, L_RETURNFLAG, char, sizeof(char));
      EXPAND_BUFFER(linestatus, L_LINESTATUS, char, sizeof(char));
      EXPAND_BUFFER(shipdate, L_SHIPDATE, uint32_t, sizeof(uint32_t));
      EXPAND_BUFFER(commitdate, L_COMMITDATE, uint32_t, sizeof(uint32_t));
      EXPAND_BUFFER(receiptdate, L_RECEIPTDATE, uint32_t, sizeof(uint32_t));
      EXPAND_BUFFER(shipinstruct, L_SHIPINSTRUCT, char, 26); 
      EXPAND_BUFFER(shipmode, L_SHIPMODE, char, 11); 
      EXPAND_BUFFER(comment, L_COMMENT, char, 45); 

      /* read next text line */
      str = fgets(buffer, sizeof(buffer), f); 
   }

   fclose(f);
   *numrows = count;
   return 0;
}

